from __future__ import annotations

from zen_core.logging.printer import pad


class TeamConfig:
    def __init__(self, members, admins, repos):
        self._members = members
        self._admins = admins
        self._repos = repos

    def members(self):
        return self._members

    def admins(self):
        return self._admins

    def repos(self):
        return self._repos

    def diff(self, older_config: TeamConfig):
        new_members = [m for m in self.members() if m not in older_config.members() and m not in older_config.admins()]
        new_admins = [m for m in self.admins() if m not in older_config.admins()]

        new_repos = {r: permission for r, permission in self.repos().items() if permission != older_config.repos().get(r, None)}

        return TeamConfig(new_members, new_admins, new_repos)

    def length(self):
        return len(self.members()) + len(self.admins()) + len(self.repos())

    def __repr__(self):
        output = []
        if self.members():
            output += [pad('Members:', 1)]
            output += [pad(member, 2) for member in self.members()]

        if self.admins():
            output += [pad('Admins:', 1)]
            output += [pad(admin, 2) for admin in self.admins()]

        repo_configs = self.repos()

        if repo_configs:
            output += [pad('Repositories:', 1)]
            output += [pad(f'{repo_name} : {permission}', 2) for repo_name, permission in repo_configs.items()]
        return '\n'.join(output)

    def __eq__(self, other):
        return self.members() == other.members() and self.admins() == other.admins() and self.repos() == other.repos()



import click


def error_out(error_message):
    return click.style(error_message, fg='red')


def focus_out(focus_message):
    return click.style(focus_message, fg='cyan')


def write(text, indent=0):
    click.echo(f'{pad(text, indent)}')


def pad(text, indent=0):
    prefix = ''.join(['   '] * indent)
    return f'{prefix}{text}'


def section(header):
    pre = '-' * 20 + f' {header} '
    remaining_length = 80 - len(pre)
    return f"{pre}{'-' * remaining_length}"



import re

from github3 import login, enterprise_login

from zen_core.configuration.config_reader import read_config
from zen_core.handlers.git_org import GitOrg
from zen_core.handlers.git_repo import GitRepo
from zen_core.handlers.git_user import GitUser

git_client = None


class GitClient:
    def __init__(self, client=None):
        if client is None:
            self._git_client = _connect()
        else:
            self._git_client = client

        self._orgs = None
        self._repos = None

    def me(self):
        return GitUser(self._git_client.me())

    def search_orgs(self, query):
        if self._orgs is None:
            self._orgs = self._git_client.organizations()

        regex = re.compile(query)
        matching_orgs = [GitOrg(org) for org in self._orgs if regex.match(org.login)]
        return matching_orgs

    def search_repos(self, query):
        if self._repos is None:
            self._repos = self._git_client.repositories()

        regex = re.compile(query)

        matching_repos = [GitRepo(repo) for repo in self._repos if regex.match(repo.full_name)]

        if not matching_repos:
            matching_repos = [GitRepo(repo) for repo in self._git_client.search_repositories(query)]

        return matching_repos

    def organization(self, org_login):
        org = self._git_client.organization(org_login)
        if org is not None:
            return GitOrg(org)
        return None


def _connect():
    global git_client
    if not git_client:
        git_credentials = read_config()
        github_url, github_token = git_credentials[0]
        if github_url == 'github.com':
            git_client = login(token=github_token)
        else:
            git_client = enterprise_login(url=f'https://{github_url}', token=github_token)
    return git_client



from zen_core.logging.printer import write, focus_out


def apply_configuration(git_client, org_configurations, dry_run):
    for org_pattern, org_config in org_configurations.items():
        matching_orgs = git_client.search_orgs(org_pattern)

        for org in matching_orgs:
            org_config = org_config.resolve(org)
            current_configuration = org.configuration()

            difference = org_config.diff(current_configuration)

            if difference.length() != 0:
                write(f'Configuration analysis for {focus_out(org.login())} is complete.')
                write(difference)

            if not dry_run:
                for m in difference.members():
                    org.grant_access(m)

                for m in difference.admins():
                    org.grant_access(m, role='admin')

                for team_name, team_config in difference.teams().items():
                    existing_team = org.team(team_name)
                    if existing_team is None:
                        existing_team = org.create_team(team_name)

                    for repo, permission in team_config.repos().items():
                        existing_team.add_to_repo(repo, permission)

                    for member in team_config.members():
                        existing_team.grant_access(member)
                    for admin in team_config.admins():
                        existing_team.grant_access(admin, role='maintainer')



import configparser
import os

import click

CONFIG_FILE_PATH = os.path.expanduser('~/.config/sentry/sentry.ini')


def read_config():
    token = os.environ.get('GIT_TOKEN', default=None)
    git_instance = os.environ.get('GIT_URL', default=None)
    if token is None or git_instance is None:
        config_parser = configparser.ConfigParser()
        if not os.path.isfile(CONFIG_FILE_PATH):
            if not os.path.isdir(os.path.dirname(CONFIG_FILE_PATH)):
                os.makedirs(os.path.dirname(CONFIG_FILE_PATH))
            github_instance = click.prompt('Github instance URL', default='github.com')
            token = click.prompt('Token', hide_input=True)
            config_parser[github_instance] = {'token': token}
            with open(CONFIG_FILE_PATH, 'w') as config_file:
                config_parser.write(config_file)

        config_parser.read(CONFIG_FILE_PATH, encoding='utf-8')
        config = [(section, config_parser[section]['token']) for section in config_parser.sections()]
        git_instance, token = config[0]

    return [(git_instance, token)]



from __future__ import annotations

import re

from zen_core.logging.printer import section, pad
from zen_core.parsing.team_config import TeamConfig


class OrgConfig:
    def __init__(self, pattern, members, admins, teams):
        self._pattern = pattern
        self._members = members
        self._admins = admins

        self._teams = teams

    def members(self):
        return self._members

    def admins(self):
        return self._admins

    def teams(self):
        return self._teams

    def diff(self, older_config: OrgConfig):
        new_members = [m for m in self.members() if m not in older_config.members() and m not in older_config.admins()]
        new_admins = [m for m in self.admins() if m not in older_config.admins()]

        team_diff = {}
        for team_name, team_config in self.teams().items():
            if team_name not in older_config.teams().keys():
                team_diff[team_name] = team_config
            else:
                old_team_config = older_config.teams()[team_name]
                new_team_config = team_config.diff(old_team_config)
                if new_team_config.length() != 0:
                    team_diff[team_name] = new_team_config
        diff = OrgConfig(self._pattern, new_members, new_admins, team_diff)
        return diff

    def resolve(self, actual_org):
        teams_with_resolved_repositories = {}
        for team_name, team_config in self.teams().items():
            resolved_members = team_config.members()
            resolved_admins = team_config.admins()
            resolved_repos = {}
            repository_patterns = team_config.repos()

            for repo_pattern, permission in repository_patterns.items():
                regex = re.compile(repo_pattern)
                for repo in actual_org.repositories():
                    if regex.match(repo.login()):
                        resolved_repos[repo.full_name()] = permission
            teams_with_resolved_repositories[team_name] = TeamConfig(resolved_members, resolved_admins, resolved_repos)

        return OrgConfig(actual_org.login(), self._members, self._admins, teams_with_resolved_repositories)

    def length(self):
        return len(self._members) + len(self._admins) + len(self._teams)

    def __repr__(self):
        output = []
        output += [section(f'Org Config for {self._pattern}')]

        if self._members:
            output += [section('Org Members')]
            output += [pad(member, 2) for member in self._members]

        if self._admins:
            output += [section('Org Owners')]
            output += [pad(admin, 2) for admin in self._admins]

        for team, team_config in self.teams().items():
            output += [f'Team: {team}']
            output += [f'{team_config}']

        return '\n'.join(output)

    def __eq__(self, other):
        if isinstance(other, OrgConfig):
            return self.members() == other.members() and self.admins() == other.admins()

        if isinstance(other, dict):
            return self.members() == other.get('members', []) and self.admins() == other.get('admins', [])



import glob
import os
import typing
from collections import OrderedDict

import toml

from zen_core.logging.printer import write
from zen_core.parsing.group_config import GroupConfig
from zen_core.parsing.org_config import OrgConfig
from zen_core.parsing.team_config import TeamConfig


def find_toml_files(path):
    return glob.glob(f'{path}/*.toml')


OrgConfigurations = typing.Dict[str, OrgConfig]


def parse_toml_configuration(path) -> OrgConfigurations:
    groups = parse_groups(path)
    orgs = _parse_orgs(path, groups)
    return orgs


def parse_groups(path):
    groups = {}
    files = find_toml_files(path)
    for f in files:
        toml_config = toml.load(f)
        groups.update(_extract_group_members(toml_config.get('group', [])))

    flattened_group_configs = {group_name: group_config.flattened(groups) for (group_name, group_config) in
                               groups.items()}
    return flattened_group_configs


def _extract_group_members(groups):
    group_configs = OrderedDict()

    for group in groups:
        group_name = group['name']
        if group_name in group_configs.keys():
            write(f'Error: Duplicate group definition {group_name}')
            exit(1)

        group_conf = GroupConfig(group_name, group.get('usernames', []), group.get('groups', []))
        group_configs[group_name] = group_conf

    return group_configs


def _parse_orgs(path, groups):
    orgs = {}
    files = find_toml_files(path)
    for f in files:
        toml_config = toml.load(f)
        orgs.update(_extract_org_configuration(toml_config.get('org', []), groups))
    return orgs


def _flatten_members(original_config, groups):
    usernames = original_config.get('usernames', [])
    for group_name in original_config.get('groups', []):
        usernames += [user for user in groups[group_name].usernames() if user not in usernames]
    return usernames


def _extract_org_configuration(orgs, groups):
    org_configs = {}
    for org in orgs:
        pattern = org['pattern']

        if pattern in org_configs.keys():
            write(f'Error: Duplicate pattern {pattern} for org configuration')
            exit(1)

        members = _flatten_members(org.get('members', {}), groups)
        admins = _flatten_members(org.get('admins', {}), groups)

        teams = {}
        for team_config in org.get('team', []):
            flat_members = _flatten_members(team_config.get('members', {}), groups)
            flat_admins = _flatten_members(team_config.get('admins', {}), groups)
            repos = {repo_config['pattern']: repo_config['permission'] for repo_config in
                     team_config.get('repo', [])}
            configuration = TeamConfig(flat_members, flat_admins, repos)
            teams[team_config['name']] = configuration

        org_config = OrgConfig(pattern, members, admins, teams)

        org_configs[pattern] = org_config

    return org_configs


if __name__ == '__main__':
    config_path = os.path.expanduser('~/workspace/git-me-up')
    orgs = parse_toml_configuration(config_path)
    print(list(orgs.values())[0])



from github3.orgs import ShortOrganization

from zen_core.handlers.access_controlled_git_object import AccessControlledGitObject
from zen_core.handlers.git_repo import GitRepo
from zen_core.handlers.git_team import GitTeam
from zen_core.handlers.git_user import GitUser
from zen_core.parsing.org_config import OrgConfig
from zen_core.parsing.team_config import TeamConfig


class GitOrg(AccessControlledGitObject):

    def __init__(self, git_org: ShortOrganization):
        super().__init__(git_org)

    def login(self):
        return self._git_object.login

    def grant_access(self, username, role='member'):
        current_permission = self.permission_for(username)
        if current_permission != 'admin':
            self._git_object.add_or_update_membership(username, role)

    def revoke_access(self, username):
        self._git_object.remove_membership(username)

    def as_dict(self):
        return self._git_object.as_dict()

    def configuration(self):
        members = [m.login() for m in self.members('member')]
        admins = [m.login() for m in self.members('admin')]
        current_teams = self.teams()
        current_team_configuration = {}
        for team in current_teams:
            team_members = [m.login() for m in team.members(role='member')]
            team_admins = [m.login() for m in team.members(role='maintainer')]

            repos = {repo.full_name(): repo.permission_for_team(team.login()) for repo in team.repositories()}
            current_team_configuration[team.login()] = TeamConfig(team_members, team_admins, repos)

        return OrgConfig(self.login(), members, admins, current_team_configuration)

    def members(self, role=None):
        return [GitUser(m) for m in self._git_object.members(role=role)]

    def permission_for(self, username):
        members = [m.login() for m in self.members(role='member')]
        admins = [m.login() for m in self.members(role='admin')]
        if username in admins:
            return 'admin'
        elif username in members:
            return 'member'
        else:
            return None

    def create_team(self, name, repos=None, permission='pull'):
        if not repos:
            repos = []
        return GitTeam(self._git_object.create_team(name, repo_names=repos, permission=permission, privacy='closed'))

    def repositories(self):
        return [GitRepo(r) for r in self._git_object.repositories()]

    def teams(self):
        return [GitTeam(t) for t in self._git_object.teams()]

    def team(self, name):
        for t in self.teams():
            if t.name() == name:
                return t
        return None

    def __repr__(self):
        return f'org::{self.login()}'



from zen_core.handlers.named_git_object import NamedGitObject


class AccessControlledGitObject(NamedGitObject):
    def __init__(self, git_object):
        super().__init__(git_object)

    def grant_access(self, username, role='member'):
        pass

    def revoke_access(self, username):
        pass

    def members(self, role=None):
        pass

    def permission_for(self, username):
        pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['zen_core',
 'zen_core.configuration',
 'zen_core.handlers',
 'zen_core.logging',
 'zen_core.parsing']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.0.1,<9.0.0', 'github3.py>=2.0.0,<3.0.0']

setup_kwargs = {
    'name': 'zen-core',
    'version': '0.2.1',
    'description': 'Github API zen core utilities',
    'long_description': None,
    'author': 'dragosdumitrache',
    'author_email': 'dragos@afterburner.dev',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from zen_core.handlers.named_git_object import NamedGitObject


class GitUser(NamedGitObject):
    def __init__(self, git_object):
        super().__init__(git_object)

    def name(self):
        return self._git_object.name

    def full_name(self):
        return self._git_object.name

    def login(self):
        return self._git_object.login

    def get_organization_membership(self, organization):
        return organization.permission_for(self.login())

    def __str__(self):
        return f'GitUser[{self.login()}]'

    def __repr__(self):
        return f'GitUser[{self.login()}]'

    def __eq__(self, other):
        return self.login() == other.login()



from zen_core.handlers.access_controlled_git_object import AccessControlledGitObject
from zen_core.handlers.git_repo import GitRepo
from zen_core.handlers.git_user import GitUser


class GitTeam(AccessControlledGitObject):
    def __init__(self, git_object):
        super().__init__(git_object)

    def name(self):
        return self._git_object.name

    def login(self):
        return self.name()

    def as_dict(self):
        return self._git_object.as_json()

    def add_to_repo(self, repository_name, permission):
        self._git_object.add_repository(repository_name, permission)

    def repositories(self):
        return [GitRepo(r) for r in self._git_object.repositories()]

    def grant_access(self, user, role='member'):
        if self.permission_for(user) != 'maintainer':
            self._git_object.add_or_update_membership(user, role)

    def revoke_access(self, username):
        super().revoke_access(username)

    def members(self, role=None):
        return [GitUser(u) for u in self._git_object.members(role)]

    def permission_for(self, username):
        if any(m.login() == username for m in self.members('maintainer')):
            return 'maintainer'
        if any(m.login() == username for m in self.members('member')):
            return 'member'
        return None

    def __eq__(self, other):
        return self.name() == other.name()

    def __repr__(self):
        return f'GitTeam[{self.name()}]'

    def __str__(self):
        return f'GitTeam[{self.name()}]'



from __future__ import annotations


class NamedGitObject:
    def __init__(self, git_object):
        self._git_object = git_object

    def name(self):
        pass

    def full_name(self):
        pass

    def login(self):
        pass

    def raw_object(self):
        return self._git_object

    def __eq__(self, other: NamedGitObject) -> bool:
        return self.login() == other.login()


from zen_core.handlers.access_controlled_git_object import AccessControlledGitObject


class GitRepo(AccessControlledGitObject):
    def __init__(self, git_object):
        super().__init__(git_object)

    def name(self):
        return self._git_object.name

    def full_name(self):
        return self._git_object.full_name

    def login(self):
        return self.name()

    def as_dict(self):
        return self._git_object.as_dict()

    def permission_for_team(self, team):
        permissions = self.as_dict().get('permissions', {})
        if permissions.get('admin', False):
            return 'admin'
        elif permissions.get('push', False):
            return 'push'
        elif permissions.get('pull', False):
            return 'pull'
        else:
            return None

    def raw_object(self):
        return self._git_object

    def owner(self):
        return self._git_object.owner

    def __eq__(self, other):
        return self.login() == other.login()

    def __repr__(self):
        return f'GitRepo[{self.full_name()}]'

    def __str__(self):
        return f'GitRepo[{self.full_name()}]'



from __future__ import annotations

from zen_core.logging.printer import section


class GroupConfig:
    def __init__(self, name, usernames, groups):
        self._name = name
        self._usernames = usernames
        self._groups = groups

    def flattened(self, all_groups_configs):
        for referenced_group in self._groups:
            referenced_group_config = all_groups_configs[referenced_group]
            flattened_referenced_group_config = referenced_group_config.flattened(all_groups_configs)
            while flattened_referenced_group_config.usernames() != referenced_group_config.usernames():
                flattened_referenced_group_config = referenced_group_config.flattened(all_groups_configs)
            all_groups_configs[referenced_group] = flattened_referenced_group_config

        for referenced_group in self._groups:
            self._usernames += [user for user in all_groups_configs[referenced_group].usernames() if
                                user not in self.usernames()]

        return self

    def diff(self, older_group: GroupConfig):
        username_additions = [user for user in self.usernames() if user not in older_group.usernames()]
        group_additions = [group for group in self.groups() if group not in older_group.groups()]
        return GroupConfig(self.name(), username_additions, group_additions)

    def name(self):
        return self._name

    def usernames(self):
        return self._usernames

    def groups(self):
        return self._groups

    def __repr__(self):
        output = []
        output += [section(f'Group configuration for {self.name()}')]
        for username in self.usernames():
            output += [username]
        return '\n'.join(output)

    def __eq__(self, other):
        if isinstance(other, GroupConfig):
            return self.usernames() == other.usernames()
        if isinstance(other, list):
            return self.usernames() == other
